package core

import (
	"fmt"
	"math"

	"git.urbach.dev/cli/q/src/asm"
	"git.urbach.dev/cli/q/src/errors"
	"git.urbach.dev/cli/q/src/eval"
	"git.urbach.dev/cli/q/src/expression"
	"git.urbach.dev/cli/q/src/types"
)

// EvaluateArray evaluates an array access.
func (f *Function) EvaluateArray(expr *expression.Expression) (*eval.Memory, error) {
	name := expr.Children[0].Token.Text(f.File.Bytes)
	base := f.VariableByName(name)

	if base == nil {
		return nil, errors.New(&errors.UnknownIdentifier{Name: name}, f.File, expr.Children[0].Token.Position)
	}

	defer f.UseVariable(base)

	memory := asm.Memory{
		Base:           base.Value.Register,
		Offset:         0,
		OffsetRegister: math.MaxUint8,
		Length:         byte(1),
	}

	indexExpr := expr.Children[1]
	index, err := f.Evaluate(indexExpr)

	if err != nil {
		return nil, err
	}

	if !types.Is(index.Type(), types.AnyInt) {
		return nil, errors.New(&errors.TypeMismatch{Encountered: index.Type().Name(), Expected: types.AnyInt.Name()}, f.File, indexExpr.Token.Position)
	}

	switch index := index.(type) {
	case *eval.Number:
		memory.Offset = int8(index.Number)
	case *eval.Register:
		memory.OffsetRegister = index.Register
	default:
		panic(fmt.Errorf("%s: not implemented: %v", f.UniqueName, index))
	}

	array, isArray := base.Value.Typ.(*types.Array)

	if isArray {
		value := &eval.Memory{
			Typ:    array.Of,
			Memory: memory,
		}

		return value, nil
	}

	pointer, isPointer := base.Value.Typ.(*types.Pointer)

	if isPointer {
		value := &eval.Memory{
			Typ:    pointer.To,
			Memory: memory,
		}

		return value, nil
	}

	panic("invalid type")
}
